/*
 * powernowd.c: (c) 2003-2004 John Clemens <clemej@alum.rpi.edu>
 *     Portions (c) 2004 Hans Ulrich Niedermann <kdev@n-dimensional.de>
 *
 * Released under the GNU General Public License. See the LICENSE file
 * included with this file.
 *
 * * Changelog:
 *   v0.75, initial release.
 *   v0.80, add syslog support, -s,-p,-u,-l options, clean up error reporting
 *          some (still some more work needed), made packaging better, removed
 *          find_mount() code.. /sys is the blessed place, so be it.. no sense
 *          having unnecessary code and the maintenance thereof.  fixed bug
 *          where mode was actually defaulting to SINE, not AGGRESSIVE.
 *   v0.85, Minor memory init fixes, add clearer error messages, check for
 *	    root, added -b and -n options, add pause support (SIGUSR1/2, -b
 *	    option), added #ifdef'd-out code to handle buggy athlons which
 *	    might not be necessary anymore, fixed some help text bugs, added
 *	    a few more comments.
 *   v0.90, Support drivers that report speed in Mhz instead of Khz. Removed
 *	    buggy athlon workaround, as it's not needed. Removed pause/unpause
 *	    code. Added sample powernowd init script to show how to emulate
 *	    old pause/unpause behavior a cleaner way. Added LEAPS mode and
 *	    verbosity patches by Hans Ulrich Niedermann. Cleaned up verbosity
 *	    handling more.
 *   newer: Added support for Hyper-Threading/SMP systems. Various buffer size
 *          and integer signedness fixes.
 *
 * * Contributions from:
 *	Warren Togami <warren@togami.com>
 *	Michael Schreiter <Michael_Schreiter@gmx.de>
 *	Wolfgang Tremmel <tremmel@garf.de>
 *	Hans Ulrich Niedermann <kdev@n-dimensional.de>
 */


#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <stdarg.h>
#include <sys/syslog.h>

#include "config.h"

#ifdef VERBOSE_BUILD_INFO
#include "package-version.h"
#include "build-info.h"
#endif

/* Define this if you're using the broken longhaul driver which
 * reports frequencies in MHz unlike all other drivers which report
 * frequencies in kHz
 */
#undef HACK_LONGHAUL

typedef struct cpustats {
	unsigned long long user;
	unsigned long long mynice;
	unsigned long long system;
	unsigned long long idle;
} cpustats_t;

typedef struct phys_cpuinfo {
	unsigned int phys_cpuid;
	unsigned int virt_cpu_count;
	unsigned int virt_cpuids[2];
	unsigned int max_speed;
	unsigned int min_speed;
	unsigned int current_speed;
	int fd;
	char *sysfs_dir;
	cpustats_t *curr_reading;
	cpustats_t *last_reading;
#ifdef HACK_LONGHAUL
	int in_mhz; /* 0 = speed in kHz, 1 = speed in mHz */
#endif
} phys_cpuinfo_t;

/*
 * Global pointer to the beginning of all our info.  This is so we can tear
 * stuff down later in the signal handler.
 */

int phys_ncpus = 0;
phys_cpuinfo_t *phys_cpuinfo;

/* idea stolen from procps */
static char file_buf[1024];

enum function {
	SINE,
	AGGRESSIVE,
	PASSIVE,
	LEAPS
} func = AGGRESSIVE;

enum modes {
	RAISE,
	LOWER
};

/* for a daemon as simple as this, global data is ok. */
/* settings */
int daemonize = 1;
int ignore_nice = 1;
int verbosity = 0;
unsigned int step = 100000;  /* in kHz */
unsigned int poll = 1000; /* in msecs */
unsigned int highwater = 80;
unsigned int lowwater = 20;

/* statistics */
unsigned int change_speed_count = 0;
time_t start_time = 0;

#define PROCFS_CPUINFO "/proc/cpuinfo"

#define SYSFS_TREE "/sys/devices/system/cpu/"
#define SYSFS_SETSPEED "scaling_setspeed"
#define SYSFS_GOVERNOR "scaling_governor"

inline static void pprintf(const int level, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	if (level <= verbosity) {
		if (daemonize)
			vsyslog(LOG_INFO, format, ap);
		else
			vprintf(format, ap);
	}
	va_end(ap);
}

inline static void pperror(const char *format, ...)
{
	int err = errno;
	va_list ap;
	va_start(ap, format);
	if (daemonize) {
		vsyslog(LOG_INFO, format, ap);
		vsyslog(LOG_INFO, ": %s", strerror(err));
	} else {
		vfprintf(stderr, format, ap);
		fprintf(stderr, ": %s\n", strerror(err)); // not thread safe!
	}
	va_end(ap);
}

static void help(void)
{

	printf("PowerNow Daemon v%s, (c) 2003-2004 John Clemens\n", VERSION);
	printf("Daemon to control the speed and voltage of cpus.\n\n");
#ifdef VERBOSE_BUILD_INFO
	printf("Revision:\n  %s\n"
	       "Build date:\n  %s\n"
	       "Build user:\n  %s\n"
	       "\n",
	       TLA_REVISION,
	       BUILD_DATE,
	       BUILD_USER);
#endif
	printf("This is a simple client to the CPUFreq driver, and uses\n");
	printf("linux kernel v2.5 sysfs interface.  You need a supported\n");
	printf("cpu, and a kernel that supports sysfs to run this daemon.\n");
	printf("\nAvailable Options:\n");
	printf("	-h	Print this help message\n");
	printf("	-d	Don't detach from terminal (default is to\n");
	printf("		detach and run in the background)\n");
	printf("	-v	Increase output verbosity, can be used more than once.\n");
	printf("	-q	Quiet mode, only emergency output.\n");
	printf("	-n	Include 'nice'd processes in calculations\n");
	printf("	-m #	Modes of operation, can be 0, 1, 2, or 3:\n");
	printf("		0 = SINE, 1 = AGGRESSIVE (default),\n");
	printf("		2 = PASSIVE, 3 = LEAPS\n");
	printf("	-s #	Frequency step in kHz (default = 100000)\n");
	printf("	-p #	Polling frequency in msecs (default = 1000)\n");
	printf("	-u #	CPU usage upper limit percentage [0 .. 100, default 80]\n");
	printf("	-l #    CPU usage lower limit percentage [0 .. 100, default 20]\n");

	printf("\n");
	return;
}

/*
 * Open a file and copy it's first 1024 bytes into the global "buf".
 * Zero terminate the buffer.
 */
static int read_file(const char *file, int fd, int new)
{
	int n, err;

	if (new != 0) {
		if ((fd = open(file, O_RDONLY)) == -1) {
			err = errno;
			pperror("Couldn't open file %s", file);
			return err;
		}
	}

	lseek(fd, 0, SEEK_SET);
	if ((n = read(fd, file_buf, sizeof(file_buf)-1)) < 0) {
		err = errno;
		pperror("Couldn't read file %s", file);
		(void) close(fd);
		return err;
	}
	file_buf[n] = '\0';

	if (new)
		(void) close(fd);

	return 0;
}

/*
 * Reads /proc/stat into buf, and parses the output.
 *
 * Format of line:
 * ...
 * cpu<id> <user> <nice> <system> <idle> <iowait>
 */
static int get_stat(phys_cpuinfo_t *cpu)
{
	int err;
	unsigned int i;

	pprintf(5, "get_stat(cpu = %p)\n", cpu);
	pprintf(5, "get_stat(phys_cpuid = %d)\n", cpu->phys_cpuid);

	if ((err = read_file("/proc/stat", cpu->fd, 0)) < 0) {
		pperror("Error reading /proc/stat");
		return err;
	}

	do {
		cpustats_t *tmp;
		tmp = cpu->curr_reading;
		cpu->curr_reading = cpu->last_reading;
		cpu->last_reading = tmp;
	} while (0);

	pprintf(5, "before loop\n");

	pprintf(3, "Reading stats for physical cpu%d)\n",
		cpu->phys_cpuid);

	for (i = 0; i < cpu->virt_cpu_count; i++) {
		char *p1, *p2;
		char searchfor[16];
		cpustats_t *reading = &cpu->curr_reading[i];

		pprintf(4, "Reading stats for virtual cpu%d (physical cpu%d)\n",
			cpu->virt_cpuids[i], cpu->phys_cpuid);

		snprintf(searchfor, sizeof(searchfor),
			 "cpu%u ", cpu->virt_cpuids[i]);

		p1 = strstr(file_buf, searchfor);
		if (p1 == NULL) {
			pperror("Error parsing /proc/stat");
			return ENOENT;
		}

		p2 = p1+strlen(searchfor);

		pprintf(5, "before strtol\n");

		reading->user = strtoll(p2, &p2, 10);
		reading->mynice = strtoll(p2, &p2, 10);
		reading->system = strtoll(p2, &p2, 10);
		reading->idle = strtoll(p2, &p2, 10);
	}

	pprintf(5, "after loop\n");

	return 0;
}

/*
 * Once a decision is made, change the speed.
 */

static int change_speed(phys_cpuinfo_t *cpu, enum modes mode)
{
	int fd, len, err;
	char writestr[100];

	if (mode == RAISE) {
		if ((func == AGGRESSIVE) || (func == LEAPS)) {
			cpu->current_speed = cpu->max_speed;
		} else {
			cpu->current_speed += step;
			if (cpu->current_speed > cpu->max_speed)
				cpu->current_speed = cpu->max_speed;
		}
	} else {
		if ((func == PASSIVE) || (func == LEAPS)) {
			cpu->current_speed = cpu->min_speed;
		} else {
			cpu->current_speed -= step;
			if (cpu->current_speed < cpu->min_speed)
				cpu->current_speed = cpu->min_speed;
		}
	}

	pprintf(5,"Setting speed to %d\n", cpu->current_speed);

	change_speed_count++;

	strncpy(writestr, cpu->sysfs_dir, 50);
	strncat(writestr, SYSFS_SETSPEED, 20);

	if ((fd = open(writestr, O_WRONLY)) < 0) {
		err = errno;
		pperror("Can't open scaling_setspeed");
		return err;
	}

	lseek(fd, 0, SEEK_CUR);

	snprintf(writestr, sizeof(writestr), "%u\n",
#ifdef HACK_LONGHAUL
		(cpu->in_mhz) ?	(cpu->current_speed / 1000) :
#endif
		cpu->current_speed);

	pprintf(4,"mode=%d, str=%s", mode, writestr);

	if ((len = write(fd, writestr, strlen(writestr))) < 0) {
		err = errno;
		pperror("Couldn't write to scaling_setspeed\n");
		return err;
	}

	if (len != strlen(writestr)) {
		printf("Could not write scaling_setspeed\n");
		close(fd);
		return EPIPE;
	}
	close(fd);

	return 0;
}

/*
 * The heart of the program... decide to raise or lower the speed.
 */
static int inline decide_speed(phys_cpuinfo_t *cpu)
{
	int err;
	unsigned int i;
	float pct;
	unsigned long long max_usage, max_total;

	pprintf(5, "decide_speed()\n");
	pprintf(5, "decide_speed(cpu = %p)\n", cpu);
	pprintf(5, "decide_speed(phys_cpuid = %d)\n", cpu->phys_cpuid);

	if ((err = get_stat(cpu)) < 0) {
		pperror("Can't get stats");
		return err;
	}

	max_usage = 0;
	max_total = 0;
	for (i = 0; i < cpu->virt_cpu_count; i++) {
		cpustats_t *curr_reading = &cpu->curr_reading[i];
		cpustats_t *last_reading = &cpu->last_reading[i];
		unsigned long long usage, total;

		total = (curr_reading->user - last_reading->user) +
			(curr_reading->system - last_reading->system) +
			(curr_reading->mynice - last_reading->mynice) +
			(curr_reading->idle - last_reading->idle);

		if (ignore_nice) {
			usage = (curr_reading->user - last_reading->user) +
				(curr_reading->system - last_reading->system);
		} else {
			usage = (curr_reading->user - last_reading->user) +
				(curr_reading->mynice - last_reading->mynice) +
				(curr_reading->system - last_reading->system);
		}

		if (usage > max_usage) {
			max_usage = usage;
			max_total = total;
		}
	}
	pct = ((float)max_usage)/((float)max_total);

	pprintf(4,"phys cpu%d usage %Lu, total %Lu\n", cpu->phys_cpuid, max_usage, max_total);
	pprintf(4,"phys cpu%d percentage = %f\n", cpu->phys_cpuid, pct);

	if ((pct > ((float)highwater/100.0)) &&
			(cpu->current_speed != cpu->max_speed)) {
		/* raise speed to next level */
		change_speed(cpu, RAISE);
	} else if ((pct < ((float)lowwater/100.0)) &&
			(cpu->current_speed != cpu->min_speed)) {
		/* lower speed */
		change_speed(cpu, LOWER);
	}

	return 0;
}

/*
 * Allocates and initialises the per-cpu data structures.
 */
static int get_per_cpu_info(phys_cpuinfo_t *cpu)
{
	char cpustr[100], scratch[100], tmp[10];
	int fd, err;

	pprintf(5, "get_per_cpu_info(phys_cpuid = %d)\n", cpu->phys_cpuid);

	cpu->sysfs_dir = (char *)malloc(50*sizeof(char));
	if (cpu->sysfs_dir == NULL) {
		pperror("Couldn't allocate per-cpu sysfs_dir");
		return ENOMEM;
	}
	memset(cpu->sysfs_dir, 0, (50*sizeof(char)));

	strncpy(cpu->sysfs_dir, SYSFS_TREE, 30);
	snprintf(cpustr, 27, "cpu%u/cpufreq/", cpu->phys_cpuid);
	strncat(cpu->sysfs_dir, cpustr, 20);

	strncpy(scratch, cpu->sysfs_dir, 50);
	strncat(scratch, "cpuinfo_max_freq", 18);
	if ((err = read_file(scratch, 0, 1)) < 0) {
		pperror("couldn't open max freq file");
		return err;
	}

	cpu->max_speed = strtol(file_buf, NULL, 10);

	strncpy(scratch, cpu->sysfs_dir, 50);
	strncat(scratch, "cpuinfo_min_freq", 18);

	if ((err = read_file(scratch, 0, 1)) < 0) {
		pperror("couldn't open min freq file");
		return err;
	}

	cpu->min_speed = strtol(file_buf, NULL, 10);
	cpu->current_speed = cpu->max_speed;

	strncpy(scratch, cpu->sysfs_dir, 50);
	strncat(scratch, SYSFS_GOVERNOR, 20);

	if ((err = read_file(scratch, 0, 1)) < 0) {
		pperror("couldn't open scaling_governors file");
		return err;
	}

	if (strncmp(file_buf, "userspace", 9) != 0) {
		if ((fd = open(scratch, O_RDWR)) < 0) {
			err = errno;
			pperror("couldn't open govn's file for writing");
			return err;
		}
		strncpy(tmp, "userspace\n", 11);
		if (write(fd, tmp, 11*sizeof(char)) < 0) {
			err = errno;
			pperror("Error writing file governor");
			close(fd);
			return err;
		}
		if ((err = read_file(scratch, fd, 0)) < 0) {
			pperror("Error reading back governor file");
			close(fd);
			return err;
		}
		close(fd);
		if (strncmp(file_buf, "userspace", 9) != 0) {
			pperror("Can't set to userspace governor, exiting");
			return EPIPE;
		}
	}

	/*
	 * Some cpufreq drivers (longhaul) report speeds in MHz instead
	 * of KHz.  Assume for now that any currently supported cpufreq
	 * processor will a) not be faster then 10GHz, and b) not be slower
	 * than 10MHz. Therefore, if the number for max_speed is less than
	 * 10000, assume the driver is reporting speeds in MHz, not KHz,
	 * and adjust accordingly.
	 */
#ifdef HACK_LONGHAUL
	cpu->in_mhz = 0;
	if (cpu->max_speed <= 10000) {
		cpu->in_mhz = 1;
		cpu->max_speed *= 1000;
		cpu->min_speed *= 1000;
		cpu->current_speed *= 1000;
	}
#endif

	if ((cpu->fd = open("/proc/stat", O_RDONLY)) < 0) {
		err = errno;
		pperror("can't open /proc/stat");
		return err;
	}

	if ((err = get_stat(cpu)) < 0) {
		pperror("can't read /proc/stat");
		return err;
	}

	return 0;
}

/*
 * Signal handler for SIGTERM/SIGINT... clean up after ourselves
 * Only install this signal handler *after* ncpus has been initialized!
 */
static void terminate(int signum)
{
	int i;
	time_t duration;

	for(i = phys_ncpus-1; i >= 0; i--) {
		/* force ourselves back up to the highest Mhz speed */
		func = LEAPS;
		change_speed(&phys_cpuinfo[i], RAISE);
		/* close the /proc/stat fd */
		close(phys_cpuinfo[i].fd);
		/* deallocate everything */
		free(phys_cpuinfo[i].sysfs_dir);
	}
	free(phys_cpuinfo);

	duration = time(NULL) - start_time;
	pprintf(1,"Statistics:\n");
	pprintf(1,"  %d speed changes in %d seconds\n",
			change_speed_count, (unsigned int) duration);
	pprintf(0,"PowerNow Daemon Exiting due to signal %d.\n", signum);

	closelog();

	exit(0);
}

static const char *str_func(void)
{
	const char *r;
	switch (func) {
	case SINE: r = "SINE"; break;
	case AGGRESSIVE: r = "AGGRESSIVE"; break;
	case PASSIVE: r = "PASSIVE"; break;
	case LEAPS: r = "LEAPS"; break;
	default: r = "UNKNOWN"; break;
	}
	return r;
}

/*
 * Initialize CPU data:
 * - Scan /proc/cpuinfo and extract number of physical CPUs
 * - Reserve correct number of cpuinfo_t's and initialize them
 */

static int init_cpu_data() {

	char *buf;
	ssize_t read_size;
	size_t buf_size;
	int fd;
	unsigned int *phys_cpuids;
	unsigned int *virt_to_phys;

	int i;

	/* Get upper limit for number of CPUs */
	unsigned int virt_ncpus = (unsigned int) sysconf(_SC_NPROCESSORS_CONF);
	if (virt_ncpus <= 0) {
		fprintf(stderr, "Invalid number of CPUs: %d\n", virt_ncpus);
		return -1;
	}

	/* Buffer for contents of /proc/cpuinfo */
	buf_size = 1024 * virt_ncpus;
	buf = malloc(buf_size);
	if (NULL == buf) {
		int err = errno;
		pperror("Could not allocate parse buffer");
		return err;
	}

	/* Read /proc/cpuinfo into buffer */
	fd = open(PROCFS_CPUINFO, O_RDONLY);
	read_size = read(fd, buf, buf_size);
	if (read_size < 0) {
		int err = errno;
		pperror("Couldn't read file " PROCFS_CPUINFO);
		close(fd);
		return err;
	}
	buf[read_size] = '\0';

	close(fd);

	/* Physical CPU/Virtual CPU maps*/
	phys_cpuids = malloc(sizeof(*phys_cpuids)*virt_ncpus);
	if (NULL == phys_cpuids) {
		int err = errno;
		pperror("Could not allocate phys_cpuids buffer");
		return err;
	}
	virt_to_phys = malloc(sizeof(*virt_to_phys)*virt_ncpus);
	if (NULL == virt_to_phys) {
		int err = errno;
		pperror("Could not allocate virt_to_phys buffer");
		return err;
	}

	/* Parse contents of buffer */
	do {
		unsigned int virtual_index;
		unsigned int phys_index;
		char *next = buf;

		for (virtual_index = 0; virtual_index < virt_ncpus; virtual_index++) {
			phys_cpuids[virtual_index] = 0;
			virt_to_phys[virtual_index] = 0;
		}

		for (virtual_index = 0, phys_index = 0;
		     virtual_index < virt_ncpus;
		     virtual_index++)
		{
			char *proc;
			char *phys_id;
			char *colon;
			unsigned int virt_cpu, phys_cpu;

			proc = strstr(next, "processor");
			if (NULL == proc) {
				break;
			}
			colon = strchr(proc+1,':');
			if (NULL == colon) {
				fprintf(stderr,
					"Fatal: No colon in \"processor\" line\n");
				return(-1);
			}
			virt_cpu = strtoul(colon+1, &next, 10);

			phys_id = strstr(next, "physical id");
			if (NULL == phys_id) {
				/* Non-SMP-Kernel: assume phys_cpu = virt_cpu */
				phys_cpuids[phys_index] = virt_cpu;
				phys_index++;
				virt_to_phys[virtual_index] = virt_cpu;
			} else {
				colon = strchr(phys_id+1,':');
				if (NULL == colon) {
					fprintf(stderr,
						"Fatal: No colon in \"physical id\" line\n");
					return(-1);
				}
				phys_cpu = strtoul(colon+1, &next, 10);
				pprintf(3, "Found virtual cpu %d on physical cpu %d\n",
					virt_cpu, phys_cpu);

				virt_to_phys[virtual_index] = phys_cpu;
				do {
					unsigned int n;
					int found = 0;
					for (n=0; n < virtual_index; n++) {
						if (virt_to_phys[n] == phys_cpu) {
							found = 1;
							break;
						}
					}
					if (!found) {
						phys_cpuids[phys_index] = phys_cpu;
						phys_index++;
					}
				} while (0);
			}
		}
		phys_ncpus = phys_index;
	} while (0);

	pprintf(0,"Found %d physical cpu%s and %d virtual cpu%s:\n",
		phys_ncpus, (phys_ncpus==1)?"":"s",
		virt_ncpus, (virt_ncpus==1)?"":"s");

	/* Malloc, initialise data structs */
	phys_cpuinfo = malloc(sizeof(phys_cpuinfo_t)*phys_ncpus);

	if (phys_cpuinfo == NULL) {
		pperror("Couldn't malloc phys_cpuinfo");
		return ENOMEM;
	}
	memset(phys_cpuinfo, 0, (sizeof(phys_cpuinfo_t)*phys_ncpus));

	for (i=0; i<phys_ncpus; i++) {
		int err;
		unsigned int n;

		phys_cpuinfo[i].phys_cpuid = phys_cpuids[i];
		phys_cpuinfo[i].virt_cpu_count = 0;
		pprintf(2, "  cpu%d: physical cpu%d:", i, phys_cpuids[i]);
		for (n=0; n < virt_ncpus; n++) {
			if (virt_to_phys[n] == phys_cpuids[i]) {
				phys_cpuinfo[i].virt_cpuids[phys_cpuinfo[i].virt_cpu_count] = n;
				phys_cpuinfo[i].virt_cpu_count++;
				pprintf(2, " virtual cpu%d", n);
			}
		}
		pprintf(2, "\n");

		phys_cpuinfo[i].curr_reading =
			malloc(sizeof(cpustats_t)* phys_cpuinfo[i].virt_cpu_count);
		if (phys_cpuinfo[i].curr_reading == NULL) {
			err = errno;
			pperror("No memory for curr_readings buffer");
			return err;
		}
		phys_cpuinfo[i].last_reading =
			malloc(sizeof(cpustats_t)* phys_cpuinfo[i].virt_cpu_count);
		if (phys_cpuinfo[i].last_reading == NULL) {
			err = errno;
			pperror("No memory for last_readings buffer");
			return err;
		}
		memset(phys_cpuinfo[i].curr_reading, 0, sizeof(cpustats_t)* phys_cpuinfo[i].virt_cpu_count);
		memset(phys_cpuinfo[i].last_reading, 0, sizeof(cpustats_t)* phys_cpuinfo[i].virt_cpu_count);

		if ((err = get_per_cpu_info(&phys_cpuinfo[i])) != 0) {
			pperror("Couldn't get per-cpu data");
			return(err);
		}
		pprintf(0,"cpu%d: %d kHz - %d kHz\n",
			phys_cpuinfo[i].phys_cpuid,
			phys_cpuinfo[i].min_speed,
			phys_cpuinfo[i].max_speed);
	}

	return 0;
}

static int parse_command_line(int argc, char **argv)
{
	while(1) {
		int c;
		long int tmp;

		c = getopt(argc, argv, "dnvqm:s:p:u:l:h");
		if (c == -1)
			break;

		switch(c) {
			case 'd':
				daemonize = 0;
				break;
			case 'n':
				ignore_nice = 0;
				break;
			case 'v':
				verbosity++;
				if (verbosity > 10) verbosity = 10;
				break;
			case 'q':
				verbosity = -1;
				break;
			case 'm':
				tmp = strtol(optarg, NULL, 10);
				if ((tmp < 0) || (tmp > 3)) {
					printf("Invalid mode specified");
					help();
					exit(ENOTSUP);
				}
				func = tmp;
				pprintf(2,"Using %s mode.\n", str_func());
				break;
			case 's':
				tmp = strtol(optarg, NULL, 10);
				if (tmp < 0) {
					printf("step must be non-negative");
					help();
					exit(ENOTSUP);
				}
				step = (unsigned int) tmp;
				pprintf(2,"Using %dHz step.\n", step);
				break;
			case 'p':
				tmp = strtol(optarg, NULL, 10);
				if (tmp < 0) {
					printf("poll must be non-negative");
					help();
					exit(ENOTSUP);
				}
				poll = (unsigned int) tmp;
				pprintf(2,"Polling every %d msecs\n", poll);
				break;
			case 'u':
				tmp = strtol(optarg, NULL, 10);
				if ((tmp < 0) || (tmp > 100)) {
					printf("upper limit must be between 0 and 100\n");
					help();
					exit(ENOTSUP);
				}
				highwater = (unsigned int) tmp;
				pprintf(2,"Using upper limit of %d%%\n",highwater);
				break;
			case 'l':
				tmp = strtol(optarg, NULL, 10);
				if ((tmp < 0) || (tmp > 100)) {
					printf("lower limit must be between 0 and 100\n");
					help();
					exit(ENOTSUP);
				}
				lowwater = (unsigned int) tmp;
				pprintf(2,"Using lower limit of %d%%\n",lowwater);
				break;
			case 'h':
			default:
				help();
				exit(0);
		}
	}

	/* one last thing to check... */
	if (lowwater > highwater) {
		printf("Invalid: lower limit higher then upper limit!\n");
		help();
		return(ENOTSUP);
	}

	return(0);
}

/*
 * Main program loop.. parse arguments, sanity chacks, setup signal handlers
 * and then enter main loop
 */
int main(int argc, char **argv)
{
	int err;

	/* parse command line parameters */
	if (0!= (err = parse_command_line(argc, argv))) {
		return(err);
	}

	/* so we don't interfere with anything, including ourself */
	nice(5);

	if (daemonize)
		openlog(PACKAGE, LOG_AUTHPRIV|LOG_PERROR, LOG_DAEMON);

	/* My ego's pretty big... */
	pprintf(0,"PowerNow Daemon v%s, (c) 2003-2004 John Clemens\n",
			VERSION);

	/* are we root?? */
	if (getuid() != 0) {
		printf("Go away, you are not root. Only root can run me.\n");
		exit(EPERM);
	}

	pprintf(1,"Settings:\n");
	pprintf(1,"  verbosity:     %4d\n", verbosity);
	pprintf(1,"  mode:          %4d     (%s)\n", func, str_func());
	pprintf(1,"  step:          %4d MHz (%d kHz)\n", step/1000, step);
	pprintf(1,"  lowwater:      %4d %%\n", lowwater);
	pprintf(1,"  highwater:     %4d %%\n", highwater);
	pprintf(1,"  poll interval: %4d ms\n", poll);

	if ((err = init_cpu_data()) != 0) {
		printf(PACKAGE " encountered an error and could not start.\n");
		printf("Please make sure that:\n");
		printf(" - You are running a v2.5/v2.6 kernel or later\n");
		printf(" - That you have sysfs mounted /sys\n");
		printf(" - That you have the core cpufreq and cpufreq-userspace\n");
		printf("   modules loaded into your kernel\n");
		printf(" - That you have the cpufreq driver for your cpu loaded,\n");
		printf("   and that it works. (check dmesg for errors)\n");
		printf("If all of the above are true, and you still have problems,\n");
		printf("please email the author: clemej@alum.rpi.edu\n");

		/* We are lazy and let the OS free our reserved memory for us... */
		return(err);
	}

	pprintf(4, "Setting up signal handlers\n");

	/* now that everything's all set up, lets set up a exit handler */
	signal(SIGTERM, terminate);
	signal(SIGINT, terminate);

	pprintf(4, "Daemonizing\n");

	if (daemonize)
		daemon(0, 0);

	start_time = time(NULL);

	pprintf(4, "Running daemon loop\n");

	while(1) {
		int i;
		usleep(poll*1000);
		for(i=0; i<phys_ncpus; i++) {
			pprintf(4, "Before decide_speed\n");
			decide_speed(&phys_cpuinfo[i]);
		}
	}
}

/*
 * Local Variables:
 * c-file-style:"linux"
 * indent-tabs-mode:t
 * End:
 */
